from typing import Union, Dict

from aiohttp import ClientResponse, ClientSession

from scrapio.structures.proxies import AbstractProxyManager


async def get_proxy_from_manager(proxy_manager: Union[AbstractProxyManager, None]) -> Union[str, None]:
    if proxy_manager:
        return await proxy_manager.get_proxy()
    return None


def create_client_session(headers: Union[None, Dict[str, str]], timeout=30) -> ClientSession:
    d_headers = {'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
               'Accept-Encoding': 'gzip, deflate, br', 'Accept-Language': 'en-US,en;q=0.9',
               'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.106 Safari/537.36'}
    if headers:
        for k, v in headers.items():
            d_headers[k] = v
    return ClientSession(headers=d_headers, conn_timeout=timeout)


def response_to_html(response: ClientResponse) -> str:
    return response._body.decode('utf-8', errors='ignore')



import logging
from typing import Union

from aiohttp import ClientSession, ClientResponse, ClientError, ClientTimeout

from scrapio.structures.proxies import AbstractProxyManager
from scrapio.utils.helpers import get_proxy_from_manager


async def get_with_client(client: ClientSession, client_timeout: ClientTimeout, proxy_manager: Union[None, AbstractProxyManager]
                          , url: str) -> ClientResponse:
    proxy = await get_proxy_from_manager(proxy_manager)
    async with client.get(url, timeout=client_timeout, proxy=proxy) as resp:
        try:
            await resp.read()
        except ClientError:
            logger = logging.getLogger('ScrapIO')
            logger.warning('ClientError for URL: {}'.format(url))
        except Exception as e:
            logger = logging.getLogger('ScrapIO')
            logger.warning('Unexpected error for URL: {}'.format(e))
        else:
            return resp


async def get_with_splash(client: ClientSession, client_timeout: ClientTimeout, url: str) -> ClientResponse:
    async with client.get(url, timeout=client_timeout) as resp:
        try:
            await resp.read()
        except ClientError:
            logger = logging.getLogger('ScrapIO')
            logger.warning('ClientError for URL: {}'.format(url))
        except Exception as e:
            logger = logging.getLogger('ScrapIO')
            logger.warning('Unexpected error for URL: {}'.format(e))
        else:
            return resp



from typing import List
from urllib.parse import urlparse, urljoin, urldefrag

from aiohttp import ClientResponse
import lxml.html as lh

from scrapio.structures.filtering import AbstractURLFilter, URLFilter
from scrapio.parsing.valid_url import valid_url


def link_extractor(response: ClientResponse, url_filter: URLFilter, defrag: bool) -> (str, List[str]):
    html = response._body.decode('utf-8', errors='ignore')
    req_url = response._url
    dom = lh.fromstring(html)
    response.__setattr__('dom', dom)
    response.__setattr__('html', html)
    found_urls = []
    for href in dom.xpath('//a/@href'):
        url = urljoin(str(req_url), href)
        if defrag:
            url = urldefrag(url)[0]
        netloc = urlparse(url).netloc
        can_crawl = url_filter.can_crawl(netloc, url)
        if can_crawl and valid_url(url):
            found_urls.append(url)
    return response, found_urls



from abc import ABC, abstractmethod
import logging
from typing import List, Union, Dict
from urllib.robotparser import RobotFileParser


class AbstractURLFilter(ABC):

    @abstractmethod
    def can_crawl(self, host: str, url: str) -> bool:
        pass


class URLFilter(AbstractURLFilter):

    __slots__ = ['_net_locations', '_additional_rules', '_robots', '_robots_cache', 'defragment']

    def __init__(self, net_locations: List[str], additional_rules: Union[List[str], None], follow_robots: bool,
                 defragment: bool = True):
        self._net_locations = net_locations
        self._additional_rules = additional_rules
        self._robots = follow_robots
        self._robots_cache = self._gather_robots_files() if follow_robots else None
        self.defragment = defragment

    def _gather_robots_files(self) -> Dict[str, RobotFileParser]:
        robot_cache = dict()
        if isinstance(self._net_locations, list):
            net_locations = self._net_locations
        else:
            net_locations = [self._net_locations]
        for host in net_locations:
            try:
                parser = RobotFileParser(url='http://{}/robots.txt'.format(host))
                parser.read()
                robot_cache[host] = parser
            except Exception:
                logger = logging.getLogger('ScrapIO')
                logger.warning('Was unable to successfully download robots.txt, for: {}'.format(host))
        return robot_cache

    def can_crawl(self, host: str, url: str) -> bool:
        if self._additional_rules:
            value = any(i in url for i in self._additional_rules)
            if value is True:
                return False
        if self._robots:
            robots_rules = self._robots_cache.get(host)
            if robots_rules is None and host in self._net_locations:
                return True
            if robots_rules:
                return robots_rules.can_fetch('*', url) and host in self._net_locations
        return host in self._net_locations


from abc import ABC, abstractmethod
from asyncio import Queue
from typing import List


class AbstractProxyManager(ABC):

    @abstractmethod
    async def get_proxy(self) -> str:
        pass


class RoundRobinProxy(AbstractProxyManager):

    def __init__(self, proxies: List[str]):
        self._queue = Queue()
        self.__launch_queue(proxies)

    def __launch_queue(self, proxies: List[str]) -> None:
        for i in proxies:
            self._queue.put_nowait(i)

    async def get_proxy(self) -> str:
        proxy = await self._queue.get()
        await self._queue.put(proxy)
        return proxy


import logging

try:
    import motor.motor_asyncio as motor_async
except ImportError:
    ImportError("motor is required to use MongoMixin")


__all__ = [
    'MongoMixin',
]


class MongoMixin:

    def create_mongo_client(self, connection_string: str, database: str, collection: str) -> None:
        client = motor_async.AsyncIOMotorClient(connection_string)
        database = client[database]
        self.__setattr__('mongo_connection', database[collection])

    async def save_results(self, result):
        connection = self.__getattribute__('mongo_connection')
        if isinstance(result, list):
            exc_func = connection.insert_many
        elif isinstance(result, dict):
            exc_func = connection.insert_one
        else:
            raise TypeError("Result must be list of dictionary objects, BSONObject or a plain dictionary")
        try:
            await exc_func(result)
        except Exception as e:
            logger = logging.getLogger('Scraper')
            logger.warning('Exception inserting document into collection: {}'.format(e))


from urllib.parse import urlparse
from typing import Union, List


def hosts_from_url(raw_urls: Union[List[str], None], start_url: str) -> List[str]:
    if raw_urls is None:
        return [urlparse(start_url).netloc]
    netlocs = []
    for url in raw_urls:
        netlocs.append(urlparse(url).netloc)
    return list(set(netlocs))


from scrapio.scrapers import BaseCrawler


class CfCrawler(BaseCrawler):

    async def _create_client_session(self):
        async with self._creation_semaphore:
            try:
                from aiocfscrape import CloudflareScraper
            except ImportError:
                raise ImportError('Cfcrawler requires aiocfscrape')
            self._client = CloudflareScraper()


import asyncio
import logging
from typing import Union, List, Any
from urllib.parse import urljoin

from aiohttp import ClientSession, ClientTimeout, ClientResponse

from scrapio.structures.filtering import URLFilter
from scrapio.parsing.links import link_extractor
from scrapio.structures.queues import NewJobQueue
from scrapio.requests.get import get_with_splash
from scrapio.structures.proxies import AbstractProxyManager
from scrapio.utils.helpers import create_client_session


__all__ = [
    'SplashConfiguration',
    'SplashCrawler',
]


class SplashConfiguration:

    def __init__(self, location: str, timeout: int, wait: int, mode: str ='render.html',):
        if mode not in ['render.html']:
            raise TypeError("SplashScraper currently only supports render.html")
        self.splash_url = self._build_splash_url(location, mode, timeout, wait)
        self._render_mode = mode

    @staticmethod
    def _build_splash_url(location, mode, timeout, wait) -> str:
        base_url = urljoin(location, mode)
        url_el = '?url={url}'
        params = '&timeout={timeout}&wait={wait}'.format(timeout=timeout, wait=wait)
        url_el += params
        return '{}{}'.format(base_url, url_el)


class SplashCrawler:

    def __init__(self, spalsh_configuration: SplashConfiguration,  start_url: Union[List[str], str],
                 max_crawl_size: Union[int, None] = None, timeout=30, user_agent=None,  **kwargs):
        self.splash_configuration = spalsh_configuration
        self._start_url = start_url
        self._client: Union[None, ClientSession] = None
        self._client_timeout: Union[None, ClientTimeout] = None
        self._timeout: Union[int, None] = None

        self._proxy_manager: \
            Union[None, AbstractProxyManager] = kwargs.get('proxy_manager')(**kwargs) if kwargs.get(
            'proxy_manager') else None

        custom_filter = kwargs.get('custom_filter')
        if custom_filter and issubclass(custom_filter, URLFilter):
            self._url_filter = custom_filter(start_url, kwargs.get('additional_rules', []), kwargs.get('follow_robots'),
                                             True)
        else:
            self._url_filter = URLFilter(start_url, kwargs.get('additional_rules', []),
                                         kwargs.get('follow_robots', True))

        self._task_queue = NewJobQueue(max_crawl_size, start_url)

        self._executor = kwargs.get('executor', None)
        logging.basicConfig(level=logging.DEBUG, format='%(message)s')
        self._logger = kwargs.get('logger', logging.getLogger("Scraper"))
        self._client_timeout = self._setup_timeout_rules(timeout)
        self.__remaining_coroutines = 0
        self.__user_agent = user_agent
        self.__creation_semaphore = asyncio.BoundedSemaphore(1)

    @staticmethod
    def _setup_timeout_rules(timeout: Union[float, int], **kwargs) -> ClientTimeout:
        if kwargs.get('client_timeout_rules') and isinstance('client_timeout_rules', dict):
            rules = kwargs.get('client_timeout_rules')
            return ClientTimeout(**rules)
        return ClientTimeout(total=float(timeout))

    async def __create_client_session(self):
        async with self.__creation_semaphore:
            if self._client is None:
                self._client = create_client_session(self.__user_agent)

    def _get_best_event_loop(self):
        try:
            import uvloop
        except ImportError:
            self._logger.info('No UVLoop reverting to base event loop implementation')
            return asyncio.get_event_loop()
        else:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            return asyncio.get_event_loop()

    def parse_result(self, response: ClientResponse) -> Any:
        raise NotImplementedError

    async def save_results(self, result):
        raise NotImplementedError

    async def __consume_request_queue(self, consumer: int):
        await self.__create_client_session()
        self.__remaining_coroutines += 1

        while True:
            try:
                item = await self._task_queue.get_next_url()
                await self._make_requests(consumer, item)
                self._task_queue.completed_task()
            except asyncio.QueueEmpty:
                self._logger.info('Coroutine: {}, No more URLs, Consumer shutting down'.format(consumer))
                self.__remaining_coroutines -= 1
                if self.__remaining_coroutines <= 0:
                    await self._client.close()
                return
            except Exception as e:
                self._logger.warning("Coroutine: {}, Encountered exception: {}".format(consumer, e))
                self._task_queue.completed_task()

    async def _make_requests(self, consumer: int, url: str):
        try:
            self._logger.info('Thread: {}, Requesting URL: {}'.format(consumer, url))
            splash_location = self.splash_configuration.splash_url.format(url=url)
            if self._proxy_manager:
                proxy = await self._proxy_manager.get_proxy()
                if proxy:
                    splash_location += '&proxy={}'.format(proxy)
            resp = await get_with_splash(self._client, self._client_timeout, splash_location)
            resp._url = url  # Splash unfortunately loses the URL redirect information from the request
            self._task_queue.put_nowait(('Parse', resp))
        except Exception as e:
            self._logger.warning("Thread: {}, Encountered exception: {}".format(consumer, e))

    async def __consume_parse_queue(self, consumer: int):
        defrag = self._url_filter.defragment
        while True:
            try:
                item = await self._task_queue.get_next_parse_job()
                await self._parse_response(consumer, item, defrag)
                self._task_queue.completed_task()
            except asyncio.QueueEmpty:
                return
            except Exception as e:
                self._logger.warning("Coroutine: {}, Encountered exception: {}".format(consumer, e))
                self._task_queue.completed_task()

    async def _parse_response(self, consumer: int, response: ClientResponse, defrag: bool) -> None:
        try:
            loop = asyncio.get_event_loop()
            html, links = await loop.run_in_executor(self._executor, link_extractor, response, self._url_filter, defrag)
            parsed_data = await loop.run_in_executor(self._executor, self.parse_result, html, response)
            await self.save_results(parsed_data)
            for link in links:
                await self._task_queue.put_unique_url(link)
        except Exception as e:
            self._logger.warning('Parser: {}, Encountered exception: {}'.format(consumer, e))

    def run_scraper(self, workers: int) -> None:
        groups = [self.__consume_parse_queue(i) for i in range(workers)]
        groups += [self.__consume_request_queue(i) for i in range(workers)]
        work_group = asyncio.gather(*groups)

        loop = self._get_best_event_loop()
        try:
            loop.run_until_complete(work_group)
        finally:
            loop.close()



import os
import sys
import pathlib
from setuptools import find_packages
from distutils.core import setup


if sys.version_info < (3, 5, 3):
    raise RuntimeError("aiohttp 3.x requires Python 3.5.3+")


here = pathlib.Path(__file__).parent


install_requires = [
    'aiohttp==3.4.4',
    'async-timeout==3.0.0',
    'attrs==18.2.0',
    'chardet==3.0.4',
    'cssselect==1.0.3',
    'idna==2.7',
    'idna-ssl==1.1.0',
    'lxml==4.2.5',
    'multidict==4.4.0',
    'yarl==1.2.6',
]


args = dict(
    name='scrapio',
    version='0.1.5.1',
    description='Aysncio web crawling framework',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],
    author='Edmund Martin',
    maintainer='Edmund Martin <edmartin101@gmail.com>',
    maintainer_email='edmartin101@gmail.com',
    url='https://github.com/EdmundMartin/Scrapio',
    packages=find_packages(exclude=('tests', '*.tests', '*.tests.*')),
    python_requires='>=3.6',
    install_requires=install_requires,
    include_package_data=True,
)

setup(**args)



from abc import ABC, abstractmethod
import asyncio
from typing import Dict
import time
from urllib.parse import urlparse


class AbstractLimiter(ABC):

    @abstractmethod
    async def limited(self, url: str) -> None:
        pass


class RateLimiter(AbstractLimiter):

    def __init__(self, reqs_per_second: int):
        self._rate_limit = reqs_per_second
        self._start_time: time.time = time.time()
        self._url_count: int = 0

    async def limited(self, url: str) -> None:
        elapsed = time.time() - self._start_time
        while self._url_count / elapsed > self._rate_limit:
            await asyncio.sleep(0.01)
            elapsed = time.time() - self._start_time
        self._url_count += 1
        return


class HostLimiter(AbstractLimiter):

    def __init__(self, host_dict: Dict[str, int]):
        self._count_per_domain: Dict[str, int] = {}
        self._hosts, self.host_times = self._load_hosts(host_dict)
        self._start_time: time.time = time.time()

    def _load_hosts(self, host_dict: Dict[str, int]):
        hosts = {}
        host_times = {}
        for k, v in host_dict.items():
            u = urlparse(k).netloc
            hosts[u] = v
            host_times[u] = v
            self._count_per_domain[u] = 0
        return hosts

    async def limited(self, url: str) -> None:
        netloc = urlparse(url).netloc
        if netloc in self._hosts:
            count: int = self._count_per_domain[netloc]
            elapsed = time.time() - self._start_time
            while count / elapsed > self._hosts[netloc]:
                await asyncio.sleep(0.01)
                elapsed = time.time() - self._start_time
            self._count_per_domain[netloc] += 1
            return
        return



import asyncio
from typing import Union, List


class NewJobQueue:

    __slots__ = ['_seen_urls', '_active_jobs', '_seen_semaphore', '_max_crawl_size', '_page_count', '_request_queue',
                 '_parse_queue']

    def __init__(self, max_crawl_size: Union[int, None], seed_urls: Union[List[str], str]):
        self._seen_urls = set()
        self._active_jobs = 0
        self._seen_semaphore = asyncio.BoundedSemaphore(1)
        self._max_crawl_size = max_crawl_size
        self._page_count = 0
        self._request_queue = asyncio.Queue()
        self._parse_queue = asyncio.Queue()
        self.seed_queue(seed_urls)

    def seed_queue(self, seed_urls: Union[List[str], str]):
        if isinstance(seed_urls, str):
            self._request_queue.put_nowait(seed_urls)
        elif isinstance(seed_urls, (set, list)):
            for item in seed_urls:
                self._request_queue.put_nowait(item)

    async def put_unique_url(self, url):
        async with self._seen_semaphore:
            if url not in self._seen_urls:
                if self._max_crawl_size and self._page_count < self._max_crawl_size:
                    await self._request_queue.put(url)
                elif self._max_crawl_size is None:
                    await self._request_queue.put(url)
                self._page_count += 1
            self._seen_urls.add(url)

    async def get_next_url(self):
        while True:
            try:
                next_job = self._request_queue.get_nowait()
                self._active_jobs += 1
            except asyncio.QueueEmpty:
                if self._active_jobs > 0:
                    await asyncio.sleep(0.001)
                elif self._parse_queue.qsize() >= 1:
                    await asyncio.sleep(0.001)
                elif self._active_jobs <= 0:
                    raise asyncio.QueueEmpty("Queue empty with no pending jobs")
                else:
                    await asyncio.sleep(0.001)
            else:
                return next_job

    async def put_parse_request(self, response):
        await self._parse_queue.put(response)

    async def get_next_parse_job(self):
        while True:
            try:
                next_job = self._parse_queue.get_nowait()
                self._active_jobs += 1
            except asyncio.QueueEmpty:
                if self._active_jobs > 0:
                    await asyncio.sleep(0.001)
                elif self._request_queue.qsize() >= 1:
                    await asyncio.sleep(0.001)
                elif self._active_jobs <= 0:
                    raise asyncio.QueueEmpty("Queue empty with no pending jobs")
                else:
                    await asyncio.sleep(0.001)
            else:
                return next_job

    def completed_task(self):
        self._active_jobs -= 1